@@ -1,8 +1,9 @@
-from test.support import verbose, is_android
+from test.support import verbose, is_android, check_warnings
 import unittest
 import locale
 import sys
 import codecs
+import warnings
 
 class BaseLocalizedTest(unittest.TestCase):
     #
@@ -143,8 +144,9 @@ class BaseFormattingTest(object):
             func(format, value, **format_opts), out)
 
     def _test_format(self, format, value, out, **format_opts):
-        self._test_formatfunc(format, value, out,
-            func=locale.format, **format_opts)
+        with check_warnings(('', DeprecationWarning)):
+            self._test_formatfunc(format, value, out,
+                func=locale.format, **format_opts)
 
     def _test_format_string(self, format, value, out, **format_opts):
         self._test_formatfunc(format, value, out,
@@ -197,6 +199,10 @@ class EnUSNumberFormatting(BaseFormattingTest):
         self._test_format("%+10.f", -4200, grouping=0, out='-4200'.rjust(10))
         self._test_format("%-10.f", 4200, grouping=0, out='4200'.ljust(10))
 
+    def test_format_deprecation(self):
+        with self.assertWarns(DeprecationWarning):
+            locale.format("%-10.f", 4200, grouping=True)
+
     def test_complex_formatting(self):
         # Spaces in formatting string
         self._test_format_string("One million is %i", 1000000, grouping=1,
@@ -227,14 +233,15 @@ class TestFormatPatternArg(unittest.TestCase):
     # Test handling of pattern argument of format
 
     def test_onlyOnePattern(self):
-        # Issue 2522: accept exactly one % pattern, and no extra chars.
-        self.assertRaises(ValueError, locale.format, "%f\n", 'foo')
-        self.assertRaises(ValueError, locale.format, "%f\r", 'foo')
-        self.assertRaises(ValueError, locale.format, "%f\r\n", 'foo')
-        self.assertRaises(ValueError, locale.format, " %f", 'foo')
-        self.assertRaises(ValueError, locale.format, "%fg", 'foo')
-        self.assertRaises(ValueError, locale.format, "%^g", 'foo')
-        self.assertRaises(ValueError, locale.format, "%f%%", 'foo')
+        with check_warnings(('', DeprecationWarning)):
+            # Issue 2522: accept exactly one % pattern, and no extra chars.
+            self.assertRaises(ValueError, locale.format, "%f\n", 'foo')
+            self.assertRaises(ValueError, locale.format, "%f\r", 'foo')
+            self.assertRaises(ValueError, locale.format, "%f\r\n", 'foo')
+            self.assertRaises(ValueError, locale.format, " %f", 'foo')
+            self.assertRaises(ValueError, locale.format, "%fg", 'foo')
+            self.assertRaises(ValueError, locale.format, "%^g", 'foo')
+            self.assertRaises(ValueError, locale.format, "%f%%", 'foo')
 
 
 class TestLocaleFormatString(unittest.TestCase):
@@ -363,9 +370,13 @@ class TestEnUSCollation(BaseLocalizedTest, TestCollation):
             raise unittest.SkipTest('wcscoll/wcsxfrm have known bugs')
         BaseLocalizedTest.setUp(self)
 
+    @unittest.skipIf(sys.platform.startswith('aix'),
+                     'bpo-29972: broken test on AIX')
     def test_strcoll_with_diacritic(self):
         self.assertLess(locale.strcoll('à', 'b'), 0)
 
+    @unittest.skipIf(sys.platform.startswith('aix'),
+                     'bpo-29972: broken test on AIX')
     def test_strxfrm_with_diacritic(self):
         self.assertLess(locale.strxfrm('à'), locale.strxfrm('b'))
 
