@@ -1,20 +1,19 @@
-import netrc, os, unittest, sys, textwrap
+import netrc, os, unittest, sys, tempfile, textwrap
+from unittest import mock
 from test import support
 
-temp_filename = support.TESTFN
 
 class NetrcTestCase(unittest.TestCase):
 
-    def tearDown(self):
-        os.unlink(temp_filename)
-
     def make_nrc(self, test_data):
         test_data = textwrap.dedent(test_data)
         mode = 'w'
         if sys.platform != 'cygwin':
             mode += 't'
-        with open(temp_filename, mode) as fp:
+        temp_fd, temp_filename = tempfile.mkstemp()
+        with os.fdopen(temp_fd, mode=mode) as fp:
             fp.write(test_data)
+        self.addCleanup(os.unlink, temp_filename)
         return netrc.netrc(temp_filename)
 
     def test_default(self):
@@ -26,6 +25,9 @@ class NetrcTestCase(unittest.TestCase):
                          ('log1', 'acct1', 'pass1'))
         self.assertEqual(nrc.hosts['default'], ('log2', None, 'pass2'))
 
+        nrc2 = self.make_nrc(nrc.__repr__())
+        self.assertEqual(nrc.hosts, nrc2.hosts)
+
     def test_macros(self):
         nrc = self.make_nrc("""\
             macdef macro1
@@ -103,8 +105,66 @@ class NetrcTestCase(unittest.TestCase):
             """, '#pass')
 
 
-def test_main():
-    support.run_unittest(NetrcTestCase)
+    @unittest.skipUnless(os.name == 'posix', 'POSIX only test')
+    def test_security(self):
+        # This test is incomplete since we are normally not run as root and
+        # therefore can't test the file ownership being wrong.
+        d = support.TESTFN
+        os.mkdir(d)
+        self.addCleanup(support.rmtree, d)
+        fn = os.path.join(d, '.netrc')
+        with open(fn, 'wt') as f:
+            f.write("""\
+                machine foo.domain.com login bar password pass
+                default login foo password pass
+                """)
+        with support.EnvironmentVarGuard() as environ:
+            environ.set('HOME', d)
+            os.chmod(fn, 0o600)
+            nrc = netrc.netrc()
+            self.assertEqual(nrc.hosts['foo.domain.com'],
+                             ('bar', None, 'pass'))
+            os.chmod(fn, 0o622)
+            self.assertRaises(netrc.NetrcParseError, netrc.netrc)
+
+    def test_file_not_found_in_home(self):
+        d = support.TESTFN
+        os.mkdir(d)
+        self.addCleanup(support.rmtree, d)
+        with support.EnvironmentVarGuard() as environ:
+            environ.set('HOME', d)
+            self.assertRaises(FileNotFoundError, netrc.netrc)
+
+    def test_file_not_found_explicit(self):
+        self.assertRaises(FileNotFoundError, netrc.netrc,
+                          file='unlikely_netrc')
+
+    def test_home_not_set(self):
+        fake_home = support.TESTFN
+        os.mkdir(fake_home)
+        self.addCleanup(support.rmtree, fake_home)
+        fake_netrc_path = os.path.join(fake_home, '.netrc')
+        with open(fake_netrc_path, 'w') as f:
+            f.write('machine foo.domain.com login bar password pass')
+        os.chmod(fake_netrc_path, 0o600)
+
+        orig_expanduser = os.path.expanduser
+        called = []
+
+        def fake_expanduser(s):
+            called.append(s)
+            with support.EnvironmentVarGuard() as environ:
+                environ.set('HOME', fake_home)
+                result = orig_expanduser(s)
+                return result
+
+        with support.swap_attr(os.path, 'expanduser', fake_expanduser):
+            nrc = netrc.netrc()
+            login, account, password = nrc.authenticators('foo.domain.com')
+            self.assertEqual(login, 'bar')
+
+        self.assertTrue(called)
+
 
 if __name__ == "__main__":
-    test_main()
+    unittest.main()
